import ast
import os
import warnings
from typing import Dict, List, Tuple, Union

import click

from righttyper.righttyper_utils import TOOL_NAME

# Suppress SyntaxWarning during AST parsing
warnings.filterwarnings("ignore", category=SyntaxWarning)

partially_annotated: List[Tuple[str, str, int]] = []
not_annotated: List[Tuple[str, str, int]] = []


class FullyQualifiedNameCollector(ast.NodeVisitor):
    def __init__(self) -> None:
        self.current_class: Union[str, None] = None
        self.qualified_names: Dict[
            ast.FunctionDef | ast.AsyncFunctionDef,
            str,
        ] = {}

    def visit_ClassDef(self, node: ast.ClassDef) -> None:
        previous_class = self.current_class
        self.current_class = node.name
        self.generic_visit(node)
        self.current_class = previous_class

    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        self._visit_function(node)

    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> None:
        self._visit_function(node)

    def _visit_function(
        self,
        node: Union[ast.FunctionDef, ast.AsyncFunctionDef],
    ) -> None:
        if self.current_class:
            qualified_name = f"{self.current_class}.{node.name}"
        else:
            qualified_name = node.name
        self.qualified_names[node] = qualified_name
        self.generic_visit(node)


def generate_fully_qualified_names_dict(
    tree: ast.AST,
) -> Dict[ast.FunctionDef | ast.AsyncFunctionDef, str]:
    collector = FullyQualifiedNameCollector()
    collector.visit(tree)
    return collector.qualified_names


def parse_python_file(
    file_path: str,
) -> List[int]:
    fully_annotated_count = 0
    partially_annotated_count = 0
    not_annotated_count = 0

    with open(file_path, "r") as file:
        tree = ast.parse(file.read(), filename=file_path)

    qualified_names = generate_fully_qualified_names_dict(tree)

    # Function to check if a node has annotations
    def has_annotations(node: ast.arg) -> bool:
        return hasattr(node, "annotation") and node.annotation is not None

    # Function to recursively search for functions in the AST
    def search_functions(node: ast.AST) -> None:
        nonlocal fully_annotated_count, partially_annotated_count, not_annotated_count

        if isinstance(node, ast.FunctionDef) or isinstance(
            node, ast.AsyncFunctionDef
        ):
            annotations_count = 0
            total_to_annotate = 1  # Count the return here, then the args

            if node.args.args:
                for index, arg in enumerate(node.args.args):
                    # Modest unsoundness:
                    # We assume that if the first argument is self or cls,
                    # it is actually in a method / classmethod
                    if index == 0 and arg.arg in ["self", "cls"]:
                        continue
                    total_to_annotate += 1
                    annotations_count += has_annotations(arg)

            # Check if there is a return type annotation
            if node.returns:
                annotations_count += 1

            entry = (
                file.name,
                qualified_names[node],
                node.lineno,
            )
            if annotations_count == total_to_annotate:
                fully_annotated_count += 1
            elif annotations_count > 0:
                partially_annotated.append(entry)
                partially_annotated_count += 1
            else:
                not_annotated.append(entry)
                not_annotated_count += 1

        for child_node in ast.iter_child_nodes(node):
            search_functions(child_node)

    search_functions(tree)

    return [
        fully_annotated_count,
        partially_annotated_count,
        not_annotated_count,
    ]


def analyze_directory(
    directory: str, cache: Dict[str, List[int]]
) -> List[int]:
    if directory in cache:
        return cache[directory]

    directory_summary: List[int] = [
        0,
        0,
        0,
    ]  # [fully annotated, partially annotated, not annotated]

    for root, dirs, files in os.walk(directory):
        if root != directory:
            # Skip processing subdirectories separately in this loop
            continue

        for file_name in files:
            if file_name.endswith(".py"):
                file_path = os.path.join(root, file_name)
                (
                    fully_annotated_count,
                    partially_annotated_count,
                    not_annotated_count,
                ) = parse_python_file(file_path)
                directory_summary[0] += fully_annotated_count
                directory_summary[1] += partially_annotated_count
                directory_summary[2] += not_annotated_count
                # Update file-level summary in cache
                cache[file_path] = [
                    fully_annotated_count,
                    partially_annotated_count,
                    not_annotated_count,
                ]

        for dir_name in dirs:
            dir_path = os.path.join(root, dir_name)
            subdir_summary = analyze_directory(dir_path, cache)
            directory_summary[0] += subdir_summary[0]  # fully annotated
            directory_summary[1] += subdir_summary[1]  # partially annotated
            directory_summary[2] += subdir_summary[2]  # not annotated

    cache[directory] = directory_summary
    return directory_summary


def analyze_all_directories(
    directory: str,
) -> Dict[str, List[int]]:
    cache: Dict[str, List[int]] = {}
    analyze_directory(directory, cache)
    return cache


def print_directory_summary(summary: Dict[str, List[int]]) -> None:
    from rich.console import Console
    from rich.table import Table

    headers = [
        "Directory",
        "Fully\nannotated",
        "Partially\nannotated",
        "Unannotated",
        "% Fully\n annotated",
    ]
    data = []
    for (
        directory,
        counts,
    ) in (
        summary.items()
    ):  # sorted(summary.items(), key=lambda x: x[1][2]):  # Sort by number of not annotated functions
        if os.path.isfile(directory):  # Check if the path is a file
            continue
        total_functions = sum(counts)
        if total_functions > 0:
            percentage_annotated = (counts[0] / total_functions) * 100
        else:
            continue
        data.append(
            [
                directory,
                f"{counts[0]}",
                f"{counts[1]}",
                f"{counts[2]}",
                f"{percentage_annotated:6.2f}%",
            ]
        )

    summary_row = data[-1]
    print(
        f"{summary_row[3]} unannotated, {summary_row[2]} partially annotated, {summary_row[1]} fully annotated ({summary_row[4]})"
    )

    table = Table(
        show_header=True,
        header_style="bold magenta",
    )
    for header in headers:
        table.add_column(header)

    for row in data:
        table.add_row(*row)

    console = Console(width=255, record=True)
    with console.capture() as _:
        console.print(table)
    html_output = console.export_html()

    # Save the HTML output to a file
    with open(f"{TOOL_NAME}-coverage.html", "w") as file:
        file.write(html_output)
    print(f"Report saved in {TOOL_NAME}-coverage.html")

    # console.print(table)


def print_file_summary(summary: Dict[str, List[int]]) -> None:
    from rich.console import Console
    from rich.table import Table

    headers = [
        "File",
        "Fully\nannotated",
        "Partially\nannotated",
        "Unannotated",
        "% Fully\nannotated",
    ]
    data = []
    # print(summary.items())
    for file_path, counts in sorted(
        summary.items()
    ):  # , key=lambda x: sum(x[1:])): # x[1][2]):  # Sort by number of not annotated functions
        if os.path.isfile(file_path):  # Check if the path is a file
            total_functions = sum(counts)
            total_annotated_functions = total_functions - counts[2] - counts[1]
            if total_functions > 0:
                percentage_annotated = (
                    total_annotated_functions / total_functions
                ) * 100
            else:
                continue
            data.append(
                [
                    file_path,
                    f"{counts[0]}",
                    f"{counts[1]}",
                    f"{counts[2]}",
                    f"{percentage_annotated:6.2f}%",
                ]
            )

    table = Table(
        show_header=True,
        header_style="bold magenta",
    )
    for header in headers:
        table.add_column(header)

    for row in data:
        table.add_row(*row)

    table.add_section()
    c = [0, 0, 0]
    for file_path, counts in summary.items():
        if not os.path.isfile(file_path):  # Check if the path is a file
            continue
        # print(f"DIRECTORY {file_path} COUNTS {counts}")
        for i, value in enumerate(counts):
            c[i] += value
    total_functions = sum(c)
    if total_functions > 0:
        percentage_annotated = (c[0] / total_functions) * 100
    else:
        percentage_annotated = 0
    summary_row = [
        "Total:",
        f"{c[0]}",
        f"{c[1]}",
        f"{c[2]}",
        f"{percentage_annotated:6.2f}%",
    ]
    table.add_row(*summary_row)

    print(
        f"{summary_row[3]} unannotated, {summary_row[2]} partially annotated, {summary_row[1]} fully annotated ({summary_row[4]})"
    )

    console = Console(width=255, record=True)
    with console.capture() as _:
        console.print(table)
    html_output = console.export_html()

    # Save the HTML output to a file
    with open(f"{TOOL_NAME}-coverage.html", "w") as file:
        file.write(html_output)
    print(f"Report saved in {TOOL_NAME}-coverage.html")


def print_annotation_summary() -> None:
    items = [partially_annotated, not_annotated]
    item_titles = [
        "Partially annotated",
        "Not annotated",
    ]
    for i in range(len(items)):
        if items[i]:
            items[i] = sorted(
                items[i],
                key=lambda x: (x[0], x[2]),
            )
            print(f"{item_titles[i]}:")
            print("-" * (len(item_titles[i]) + 1))
            while items[i]:
                fname = items[i][0][0]
                print(f"{fname}:")
                while items[i] and fname == items[i][0][0]:
                    rec = items[i].pop(0)
                    print(f"  {rec[2]:5}:{rec[1]}")
            print()


@click.command()
@click.argument("directory", type=click.Path(exists=True))
@click.option(
    "-l",
    "--level",
    type=click.Choice(["directory", "file"]),
    default="directory",
    help="Level of statistics to report: directory or file.",
)
def main(directory: str, level: str) -> None:
    if level == "directory":
        directory_summary = analyze_all_directories(directory)
        print_directory_summary(directory_summary)
    elif level == "file":
        file_summary = analyze_all_directories(directory)
        print_file_summary(file_summary)
    print_annotation_summary()


if __name__ == "__main__":
    main()



import logging
import os
import re
from functools import cache
from typing import (
    Any,
    Dict,
    Final,
    List,
    Optional,
    Set,
    Tuple,
)

from righttyper.righttyper_types import (
    ArgInfo,
    ArgumentName,
    ArgumentType,
    Filename,
    FuncInfo,
    FunctionName,
    Typename,
    TypenameSet,
)

TOOL_ID: int = 3
TOOL_NAME: Final[str] = "righttyper"
_SAMPLING_INTERVAL = 0.01
_DEBUG_PRINT: bool = False

logger = logging.getLogger("righttyper")


def reset_sampling_interval() -> None:
    global _SAMPLING_INTERVAL
    _SAMPLING_INTERVAL = 1.0


def get_sampling_interval() -> float:
    return _SAMPLING_INTERVAL


def update_sampling_interval(instrumentation_overhead, target_overhead) -> None:
    global _SAMPLING_INTERVAL
    if instrumentation_overhead < target_overhead:
        _SAMPLING_INTERVAL *= 0.9
    else:
        _SAMPLING_INTERVAL *= 1.2
    print(f"{_SAMPLING_INTERVAL=}")
    ## FIXME _SAMPLING_INTERVAL *= 1.5


def debug_print(args: Any, *varargs: Any, **kwargs: Any) -> None:
    if _DEBUG_PRINT:
        print(__file__, args, *varargs, **kwargs)


def debug_print_set_level(level: bool) -> None:
    _DEBUG_PRINT = level


@cache
def adjusted_file_name(file_name: str, start: str = "") -> str:
    return os.path.relpath(
        os.path.splitext(file_name)[0],
        start=start,
    ).replace(os.sep, ".")


def adjusted_type_name(fname: str, typename: str) -> Typename:
    # Adjust the type based on the context of the file_name
    new_typename = typename
    if typename.startswith(fname + "."):
        new_typename = typename[len(fname) + 1 :]
    return Typename(new_typename)


def unannotated(f: object, ignore_annotations: bool = False) -> Set[str]:
    """
    Returns a set of the unannotated arguments and, if
    unannotated, the return value (called "return"), for the
    given function.
    """
    import inspect

    assert inspect.isfunction(f)

    sig = inspect.signature(f)
    unannotated_args = set()

    for name, param in sig.parameters.items():
        if ignore_annotations or param.annotation is param.empty:
            unannotated_args.add(name)
    if ignore_annotations or sig.return_annotation == inspect.Signature.empty:
        unannotated_args.add("return")

    return unannotated_args


def union_typeset_str(
    file_name: str,
    typeset: TypenameSet,
    namespace: Dict[str, Any] = globals(),
    threshold_frequency: float = 0.25,
) -> Typename:
    adjusted_typeset = typeset  # trim_and_test(typeset, threshold_frequency)
    retval = None
    if not typeset:
        # Never observed any return types. Since we always sample at least once,
        # this means we did not return any values.
        retval = Typename("None")
    elif len(typeset) == 1:
        retval = adjusted_type_name(
            adjusted_file_name(file_name),
            list(typeset)[0].typename,
        )
    elif super := find_most_specific_common_superclass_by_name(
        list(
            adjusted_type_name(
                adjusted_file_name(file_name),
                t.typename,
            )
            for t in adjusted_typeset
        ),
        namespace,
    ):
        retval = super
    else:
        if adjusted_typeset:
            typenames = sorted(
                adjusted_type_name(
                    adjusted_file_name(file_name),
                    t.typename,
                )
                for t in adjusted_typeset
            )
            if len(typenames) == 1:
                return typenames[0]
            # Make a variant of typenames without Nones
            not_none_typenames = sorted(
                t.typename for t in adjusted_typeset if t.typename != "None"
            )
            # If None is one of the typenames, make the union an optional over the remaining types.
            if len(not_none_typenames) < len(typenames):
                return Typename(
                    "Optional[" + ", ".join(not_none_typenames) + "]"
                )
            if len(typenames) > 1:
                # Just Union everything.
                return Typename("Union[" + ", ".join(typenames) + "]")
            else:
                return typenames[0]
    if not retval:
        # Worst-case
        return Typename("Any")
    return retval


def find_most_specific_common_superclass_by_name(
    type_names: List[str],
    namespace: Dict[str, Any] = globals(),
) -> Optional[Typename]:
    if not type_names:
        return None

    try:
        classes = [namespace[name] for name in type_names]
    except KeyError:
        # Fail gracefully
        return None
    common_superclasses = set.intersection(
        *(set(cls.mro()) for cls in classes)
    )
    common_superclasses.discard(object)
    if not common_superclasses:
        return None

    return Typename(
        max(
            common_superclasses,
            key=lambda cls: cls.__mro__.index(object),
        ).__name__
    )


def make_type_signature(
    file_name: str,
    func_name: str,
    args: List[ArgInfo],
    retval: TypenameSet,
    namespace: Dict[str, Any],
    not_annotated: Dict[FuncInfo, Set[str]],
    arg_types: Dict[
        Tuple[FuncInfo, ArgumentName],
        ArgumentType,
    ],
    existing_annotations: Dict[FuncInfo, Dict[str, str]],
) -> str:
    # print(f"make_type_signature {file_name} {func_name} {args} {retval}")
    t = FuncInfo(
        Filename(file_name),
        FunctionName(func_name),
    )
    s = f"def {func_name}("
    for index, arginfo in enumerate(args):
        argname = arginfo.arg_name
        arg_type = arg_types[
            FuncInfo(Filename(file_name),
                     FunctionName(func_name)),
            argname,
        ]
        if arg_type == ArgumentType.vararg:
            arg_prefix = "*"
        elif arg_type == ArgumentType.kwarg:
            arg_prefix = "**"
        else:
            assert arg_type == ArgumentType.positional
            arg_prefix = ""
        if argname not in existing_annotations[t]:  # not_annotated[t]:
            argtype_fullname_set = arginfo.type_name_set
            argtype_fullname = Typename(
                union_typeset_str(
                    file_name,
                    argtype_fullname_set,
                    namespace,
                )
            )

            s += f"{arg_prefix}{argname}: {argtype_fullname}"
        else:
            if argname in existing_annotations[t]:
                s += f"{arg_prefix}{argname}: {existing_annotations[t][argname]}"
            else:
                s += f"{arg_prefix}{argname}"

        if index < len(args) - 1:
            s += ", "
    s += ")"
    if "return" in existing_annotations[t]:
        retval_name = Typename(existing_annotations[t]["return"])
    else:
        # if "return" in not_annotated[t]:
        retval_name = union_typeset_str(file_name, retval, namespace)
    s += f" -> {retval_name}:"
    return s


@cache
def skip_this_file(
    filename: str,
    script_dir: str,
    include_all: bool,
    include_files_regex: str,
) -> bool:
    debug_print(
        f"checking skip_this_file: {script_dir=}, {filename=}, {include_files_regex=}"
    )
    if include_all:
        should_skip = False
    else:
        should_skip = (
            filename.startswith("<")
            or filename.startswith("/Library")
            or filename.startswith("/opt/homebrew/")
            or "/site-packages/" in filename
            or "righttyper.py" in filename
            or script_dir not in os.path.abspath(filename)
        )
    if include_files_regex:
        should_skip = should_skip or not re.search(
            include_files_regex, filename
        )
    # print(f"result  skip_this_file {filename}: {should_skip=}")
    return should_skip



from typing import Set

import libcst as cst

from righttyper.get_import_details import (
    generate_import_nodes,
)
from righttyper.righttyper_types import ImportInfo


class ConstructImportTransformer(cst.CSTTransformer):
    def __init__(
        self,
        imports: Set[ImportInfo],
        root_path: str,
    ) -> None:
        self.imports = imports
        self.root_path = root_path

    def leave_Module(
        self,
        original_node: cst.Module,
        updated_node: cst.Module,
    ) -> cst.Module:
        new_imports = []  # set()

        # import site
        # import sysconfig

        # Paths to the main and user libraries, with an os separator added
        # Used to generate import statements later.
        # purelib = sysconfig.get_paths()["purelib"]
        # platstdlib = sysconfig.get_paths()["platstdlib"]
        # userlib = site.getusersitepackages()

        for imp in self.imports:
            q = generate_import_nodes(imp.import_details)
            new_imports.extend(q)

        # Add all import statements at the beginning of the module
        return updated_node.with_changes(
            body=list(new_imports) + list(updated_node.body)
        )



import ast
from typing import TextIO

from righttyper.righttyper_types import (
    FunctionName,
)


def generate_stub(
    input_file: str,
    output_file: str,
    exclude_functions: list[FunctionName],
) -> None:
    if exclude_functions is None:
        exclude_functions = []

    with open(input_file, "r") as file:
        tree = ast.parse(file.read())

    with open(output_file, "a") as stub_file:
        for node in tree.body:
            if isinstance(node, ast.FunctionDef) or isinstance(
                node, ast.AsyncFunctionDef
            ):
                if node.name not in exclude_functions:
                    generate_function_stub(node, stub_file)
            elif isinstance(node, ast.ClassDef):
                generate_class_stub(node, stub_file)
            elif isinstance(node, ast.Assign):
                generate_variable_stub(node, stub_file)


def generate_function_stub(
    node: ast.FunctionDef | ast.AsyncFunctionDef,
    stub_file: TextIO,
) -> None:
    args = []
    for arg in node.args.args:
        arg_str = arg.arg
        if arg.annotation:
            arg_str += f": {ast.unparse(arg.annotation).strip()}"
        args.append(arg_str)
    if node.args.vararg:
        vararg_str = f"*{node.args.vararg.arg}"
        if node.args.vararg.annotation:
            vararg_str += (
                f": {ast.unparse(node.args.vararg.annotation).strip()}"
            )
        args.append(vararg_str)
    if node.args.kwarg:
        kwarg_str = f"**{node.args.kwarg.arg}"
        if node.args.kwarg.annotation:
            kwarg_str += f": {ast.unparse(node.args.kwarg.annotation).strip()}"
        args.append(kwarg_str)
    args_str = ", ".join(args)
    if node.returns:
        returns = ast.unparse(node.returns).strip()
    else:
        returns = "None"
    stub_file.write(f"def {node.name}({args_str}) -> {returns}: ...\n")


def generate_class_stub(node: ast.ClassDef, stub_file: TextIO) -> None:
    stub_file.write(f"class {node.name}:\n")
    empty_body = True
    for elem in node.body:
        if isinstance(elem, ast.FunctionDef):
            if elem.name not in [
                "__init__",
                "__new__",
            ]:
                stub_file.write("    ")
                generate_function_stub(elem, stub_file)
                empty_body = False
        elif isinstance(elem, ast.Assign):
            stub_file.write("    ")
            generate_variable_stub(elem, stub_file)
            empty_body = False
    # FIXME TODO what about members?
    if empty_body:
        stub_file.write("    pass")
    stub_file.write("\n")


def generate_variable_stub(node: ast.Assign, stub_file: TextIO) -> None:
    for target in node.targets:
        if isinstance(target, ast.Name):
            if node.value:
                annotation = "Any"
                if isinstance(node.value, ast.Constant):
                    annotation = type(node.value.value).__name__
                stub_file.write(f"{target.id}: {annotation}\n")
            else:
                stub_file.write(f"{target.id}: Any\n")



import inspect
import logging
import sys
from functools import lru_cache
from types import ModuleType
from typing import Any, List

from righttyper.righttyper_types import ImportDetails

import libcst as cst

logger = logging.getLogger("righttyper")


# @lru_cache
def get_import_details(
    obj: Any,
) -> ImportDetails:
    """
    Get details about the import of an object, including its name, aliases, module name, and module aliases.

    Parameters:
    obj (Any): The imported object (e.g., a class, function, or module) for which to retrieve import details.

    Returns:
    ImportDetails:
        - The name of the object.
        - A set of aliases used for the object in the current global scope.
        - The name of the module from which the object was imported.
        - A set of aliases used for the module in the current global scope.

    Example:
    from rich.console import Console as Banana
    import numpy as np
    from some.thing import thingy as barnacle

    print(get_import_details(Banana))     # ImportDetails("Console", ["Banana"], "rich.console", [])
    print(get_import_details(np.ndarray)) # ImportDetails("ndarray", [], "numpy", ["np"])
    print(get_import_details(np))         # ImportDetails("numpy", ["np"], "numpy", ["np"])
    print(get_import_details(barnacle))   # ImportDetails("thingy", ["barnacle"], "some.thing", [])
    """
    obj_name = obj.__name__ if hasattr(obj, "__name__") else type(obj).__name__
    module_name = (
        obj.__module__
        if hasattr(obj, "__module__")
        else obj.__class__.__module__
    )

    # Check if the object is built-in
    if module_name == "builtins":
        return ImportDetails(
            obj_name,
            frozenset(),
            "builtins",
            frozenset(),
        )

    module_aliases = set()
    obj_aliases = set()
    frame = inspect.currentframe()
    frame = getattr(frame, "f_back", None)

    while frame:

        # Get set of function parameters which we will exclude as names
        local_scope_exclusions = set(
            frame.f_code.co_varnames[: frame.f_code.co_argcount]
        )

        # Check both globals and locals of the current frame
        for scope in [
            frame.f_globals,
            frame.f_locals,
        ]:
            module_obj = sys.modules.get(module_name)
            for (
                alias_name,
                imported_obj,
            ) in scope.items():
                if (
                    isinstance(imported_obj, ModuleType)
                    and imported_obj is module_obj
                ):
                    if (
                        alias_name != module_name
                        and alias_name not in module_aliases
                        and alias_name not in local_scope_exclusions
                    ):
                        module_aliases.add(alias_name)
                elif (
                    imported_obj is obj
                    and alias_name != obj_name
                    and alias_name not in obj_aliases
                    and alias_name not in local_scope_exclusions
                ):
                    obj_aliases.add(alias_name)

        frame = frame.f_back
    tup = ImportDetails(
        obj_name,
        frozenset(obj_aliases),
        module_name,
        frozenset(module_aliases),
    )

    return tup


def testme() -> None:
    import numpy
    import numpy as np
    from rich.console import Console as ApplePie
    from rich.console import Console as Banana

    # Test the functions
    assert get_import_details(numpy.ndarray) == ImportDetails(
        "ndarray",
        frozenset({}),
        "numpy",
        frozenset({"np"}),
    )
    assert get_import_details(np.ndarray) == ImportDetails(
        "ndarray",
        frozenset({}),
        "numpy",
        frozenset({"np"}),
    )
    assert get_import_details(Banana) == ImportDetails(
        "Console",
        frozenset({"ApplePie", "Banana"}),
        "rich.console",
        frozenset({}),
    )
    # print_possible_imports(get_import_details(Banana))
    assert get_import_details(ApplePie) == ImportDetails(
        "Console",
        frozenset({"ApplePie", "Banana"}),
        "rich.console",
        frozenset({}),
    )
    assert get_import_details(lru_cache) == ImportDetails(
        "lru_cache",
        frozenset({}),
        "functools",
        frozenset({}),
    )


def print_possible_imports(
    details: ImportDetails,
) -> None:
    """
    Print every possible import statement that might be needed based on the details provided.

    Parameters:
    details (ImportDetails):
        The output from `get_import_details` function, containing:
        - The name of the object.
        - A set of aliases used for the object in the current global scope.
        - The name of the module from which the object was imported.
        - A set of aliases used for the module in the current global scope.
    """

    # Print import statement for the object
    print(f"from {details.import_module_name} import {details.object_name}")
    for alias in details.object_aliases:
        print(
            f"from {details.import_module_name} import {details.object_name} as {alias}"
        )

    # Print import statement for the module
    print(f"import {details.import_module_name}")
    for alias in details.module_aliases:
        print(f"import {details.import_module_name} as {alias}")


# Example usage with the provided test data
# details = get_import_details(Banana)
# print_possible_imports(details)


def generate_import_nodes(
    details: ImportDetails,
) -> List[cst.Import | cst.ImportFrom | cst.EmptyLine]:
    """
    Generate import nodes for libcst based on the details provided.

    Parameters:
    details (Tuple[str, List[str], str, List[str]]):
        The output from `get_import_details` function, containing:
        - The name of the object.
        - A set of aliases used for the object in the current global scope.
        - The name of the module from which the object was imported.
        - A set of aliases used for the module in the current global scope.

    Returns:
    List[cst.CSTNode]: A list of libcst import nodes.
    """

    import_nodes: List[cst.Import | cst.ImportFrom | cst.EmptyLine] = []

    # Helper function to create a dotted name
    def create_dotted_name(
        name: str,
    ) -> cst.Attribute | cst.Name:
        parts = name.split(".")
        if len(parts) == 1:
            return cst.Name(parts[0])
        base: cst.Name | cst.Attribute = cst.Name(parts[0])
        for part in parts[1:]:
            base = cst.Attribute(value=base, attr=cst.Name(part))
        return base

    # Import statement for the object
    try:

        import_nodes.append(
            cst.ImportFrom(
                module=create_dotted_name(details.import_module_name),
                names=[cst.ImportAlias(name=cst.Name(details.object_name))],
            )
        )
        import_nodes.append(cst.EmptyLine())
    except cst.CSTValidationError:
        logger.warning(
            f"Failed to add from {details.import_module_name=} import {details.object_name=}"
        )

    for alias in details.object_aliases:
        try:
            import_nodes.append(
                cst.ImportFrom(
                    module=create_dotted_name(details.import_module_name),
                    names=[
                        cst.ImportAlias(
                            name=cst.Name(details.object_name),
                            asname=cst.AsName(name=cst.Name(alias)),
                        )
                    ],
                )
            )
            import_nodes.append(cst.EmptyLine())
        except cst.CSTValidationError:
            logger.warning(
                f"Failed to add import {details.object_name=} as {alias=}"
            )

    # Import statement for the module
    stmt = cst.Import(
        names=[
            cst.ImportAlias(
                name=create_dotted_name(details.import_module_name)
            )
        ]
    )
    import_nodes.append(stmt)
    import_nodes.append(cst.EmptyLine())

    for alias in details.module_aliases:
        import_nodes.append(
            cst.Import(
                names=[
                    cst.ImportAlias(
                        name=create_dotted_name(details.import_module_name),
                        asname=cst.AsName(name=cst.Name(alias)),
                    )
                ]
            )
        )
        import_nodes.append(cst.EmptyLine())

    return import_nodes


# Example usage with the provided test data
# details = get_import_details(Banana)
# import_nodes = generate_import_nodes(details)
# for node in import_nodes:
#    print(cst.Module([]).code_for_node(node))


if __name__ == "__main__":
    testme()



from typing import Optional

import libcst as cst


class InsertTypingImportTransformer(cst.CSTTransformer):
    def __init__(self) -> None:
        self.has_typing_import = False

    def visit_ImportFrom(self, node: cst.ImportFrom) -> Optional[bool]:
        # Check if there is already an import from 'typing' with '*'
        if (
            isinstance(node.module, cst.Name)
            and node.module.value == "typing"
            and isinstance(node.names, cst.ImportStar)
        ):
            self.has_typing_import = True
        return None

    def leave_Module(
        self,
        original_node: cst.Module,
        updated_node: cst.Module,
    ) -> cst.Module:
        if not self.has_typing_import:
            # Create the import statement
            typing_import = cst.SimpleStatementLine(
                body=[
                    cst.ImportFrom(
                        module=cst.Name(value="typing"),
                        names=cst.ImportStar(),
                    )
                ]
            )
            # Insert the import at the top of the module
            new_body = [typing_import] + list(updated_node.body)
            self.has_typing_import = (
                True  # Ensure we don't add the import more than once
            )
            return updated_node.with_changes(body=new_body)
        return updated_node



from collections import defaultdict
from typing import (
    Any,
    Dict,
    List,
    Set,
    Tuple,
)

import libcst as cst
import logging
import os
import pathlib

from righttyper.annotate_function_transformer import (
    AnnotateFunctionTransformer,
)
from righttyper.construct_import_transformer import (
    ConstructImportTransformer,
)
from righttyper.generate_stubs import (
    generate_stub,
)
from righttyper.insert_typing_import_transformer import (
    InsertTypingImportTransformer,
)
from righttyper.righttyper_types import (
    ArgInfo,
    ArgumentName,
    ArgumentType,
    Filename,
    FuncInfo,
    FunctionName,
    ImportInfo,
    Typename,
    TypenameSet,
)
from righttyper.righttyper_utils import (
    adjusted_file_name,
    adjusted_type_name,
    debug_print,
    make_type_signature,
    skip_this_file,
    union_typeset_str,
)

logger = logging.getLogger("righttyper")


def preface_with_typing_import(
    source_code: str,
) -> str:
    tree = cst.parse_module(source_code)
    transformer = InsertTypingImportTransformer()
    new_tree = tree.visit(transformer)
    return new_tree.code


def process_file(
    filename: Filename,
    type_annotations: Dict[
        FuncInfo,
        Tuple[
            List[Tuple[ArgumentName, Typename]],
            Typename,
        ],
    ],
    imports: Set[ImportInfo],
    overwrite: bool,
    not_annotated: Dict[FuncInfo, Set[str]],
    ignore_annotations: bool = False,
    srcdir: str = "",
) -> None:
    debug_print(f"process_file: {filename}")
    # print(f"process {filename}: {imports}")
    try:
        with open(filename, "r") as file:
            source = file.read()
    except FileNotFoundError:
        return

    # Make a backup
    if overwrite:
        with open(filename + ".bak", "w") as file:
            file.write(source)

    # First, update all type annotations so they are relative to the
    # source directory (--srcdir)
    for fi in type_annotations:
        adj_fname = adjusted_file_name(fi.file_name)
        args, retval_type = type_annotations[fi]
        new_arglist = []
        for arg in args:
            new_arg_type = adjusted_type_name(adj_fname, arg[1])
            new_arglist.append((arg[0], new_arg_type))
        new_retval_type = adjusted_type_name(adj_fname, retval_type)
        type_annotations[fi] = (
            new_arglist,
            new_retval_type,
        )

    # Now, rewrite all function definitions with annotations.
    cst_tree = cst.parse_module(source)
    transformer = AnnotateFunctionTransformer(
        filename, type_annotations, not_annotated
    )
    modified_tree = cst_tree.visit(transformer)

    # No changes, just skip the rest
    if modified_tree.code == source:
        return

    # Add an import statement if needed.
    # FIXME: this messes with from __future__ imports, which need to be the first import
    transformed = preface_with_typing_import(modified_tree.code)

    # If there are needed imports for class defs, process these
    needed_imports = set(
        imp for imp in imports if imp.function_fname == filename
    )
    if needed_imports:
        tree = cst.parse_module(transformed)
        import_transformer = ConstructImportTransformer(
            imports=needed_imports,
            root_path=srcdir,
        )
        try:
            transformed_tree = tree.visit(import_transformer)
            transformed = transformed_tree.code
        except Exception as e:
            import traceback

            print(traceback.format_exc())
            print(e)

    with open(
        filename + ("" if overwrite else ".typed"),
        "w",
    ) as file:
        file.write(transformed)


# Convert the collected data into the expected format for type_annotations
def collect_data(
    file_name: str,
    visited_funcs: Set[FuncInfo],
    visited_funcs_arguments: Dict[FuncInfo, List[ArgInfo]],
    visited_funcs_retval: Dict[FuncInfo, TypenameSet],
    namespace: Dict[str, Any] = globals(),
) -> Dict[
    FuncInfo,
    Tuple[
        List[Tuple[ArgumentName, Typename]],
        Typename,
    ],
]:
    type_annotations: Dict[
        FuncInfo,
        Tuple[
            List[Tuple[ArgumentName, Typename]],
            Typename,
        ],
    ] = {}
    for t in visited_funcs:
        args = visited_funcs_arguments[t]
        arg_annotations = [
            (
                ArgumentName(arginfo.arg_name),
                union_typeset_str(
                    file_name,
                    arginfo.type_name_set,
                    namespace,
                ),
            )
            for arginfo in args
        ]
        if t in visited_funcs_retval:
            retval = union_typeset_str(
                file_name,
                visited_funcs_retval[t],
                namespace,
            )
        else:
            retval = Typename("None")
        type_annotations[t] = (
            arg_annotations,
            retval,
        )
        # print(f"{type_annotations[t]} {t}")
    return type_annotations


def output_stub_files(
    namespace: Dict[str, Any],
    root_path: str,
    imports: Set[ImportInfo],
    visited_funcs: Set[FuncInfo],
    script_dir: str,
    include_all: bool,
    include_files_regex: str,
    visited_funcs_arguments: Dict[FuncInfo, List[ArgInfo]],
    visited_funcs_retval: Dict[FuncInfo, TypenameSet],
    not_annotated: Dict[FuncInfo, Set[str]],
    arg_types: Dict[
        Tuple[FuncInfo, ArgumentName],
        ArgumentType,
    ],
    existing_annotations: Dict[FuncInfo, Dict[str, str]],
) -> None:
    # Print all type signatures
    output_str: Dict[str, str] = defaultdict(str)  # map file name to str
    already_imported: Set[Tuple[str, str]] = set()
    skip_functions: Dict[str, List[FunctionName]] = defaultdict(list)

    import site
    import sysconfig

    # Paths to the main and user libraries, with an os separator added
    # Used to generate import statements later.
    purelib = sysconfig.get_paths()["purelib"]
    platstdlib = sysconfig.get_paths()["platstdlib"]
    userlib = site.getusersitepackages()

    # Precompute imports mapping
    imports_map = {}
    for imp in imports:
        if (
            imp.class_fname.startswith(purelib)
            or imp.class_fname.startswith(userlib)
            or imp.class_fname.startswith(platstdlib)
        ):
            if imp.class_fname.startswith(
                purelib
            ) or imp.class_fname.startswith(userlib):
                class_src_file = os.path.dirname(imp.class_fname)
                class_src_file = class_src_file.removeprefix(purelib + os.sep)
                class_src_file = class_src_file.removeprefix(userlib + os.sep)
            else:
                class_src_file = imp.class_fname.removeprefix(platstdlib)
                class_src_file = os.path.basename(class_src_file)
                class_src_file, _ = os.path.splitext(class_src_file)

            class_src_file = class_src_file.replace(os.sep, ".")
            imports_map[(imp.function_fname, imp.class_name)] = class_src_file
            continue
        if imp.class_fname == "":
            # Note: not sure why this is happening.
            continue
        try:
            normalized_path = os.path.relpath(imp.class_fname, start=root_path)
        except ValueError:
            logger.warning(f"ValueError: {imp.class_fname=} {root_path=}")
            continue
        module_path, _ = os.path.splitext(normalized_path)
        if module_path.startswith(".."):  # SOMETHING HAS GONE OF THE TRACKS
            module_path = os.path.abspath(module_path)
        module_path = module_path.replace(os.sep, ".")
        if module_path.endswith(".__init__"):
            module_path = module_path[:-9]  # remove '.__init__'
        assert not module_path.startswith(
            ".."
        )  # SOMETHING HAS GONE OF THE TRACKS

        imports_map[(imp.function_fname, imp.class_name)] = module_path

    for t in visited_funcs:
        if skip_this_file(
            t.file_name,
            script_dir,
            include_all,
            include_files_regex,
        ):
            continue
        try:
            stub_fname = str(pathlib.Path(t.file_name).with_suffix(".pyi"))

            # Add relevant imports
            for (
                function_file_path,
                class_name,
            ), module_path in imports_map.items():
                if (
                    stub_fname,
                    class_name,
                ) not in already_imported:
                    if not class_name.startswith("_"):
                        # We refuse to try to import names starting with underscores, at least for now.
                        output_str[
                            stub_fname
                        ] += f"from {module_path} import {class_name}\n"
                    already_imported.add((stub_fname, class_name))

            s = make_type_signature(
                file_name=t.file_name,
                func_name=t.func_name,
                args=visited_funcs_arguments[t],
                retval=visited_funcs_retval[t],
                namespace=namespace,
                not_annotated=not_annotated,
                arg_types=arg_types,
                existing_annotations=existing_annotations,
            )
            skip_functions[stub_fname].append(t.func_name)
            output_str[stub_fname] += f"{s} ...\n"

        except KeyError:
            # Something weird happened
            logger.warning(f"KeyError: {t=}")

    for (
        stub_fname,
        stub_content,
    ) in output_str.items():
        with open(stub_fname, "w") as f:
            f.write(stub_content)

        input_file = str(pathlib.Path(stub_fname).with_suffix(".py"))
        output_file = stub_fname
        exclude_functions = skip_functions[stub_fname]
        generate_stub(
            input_file,
            output_file,
            exclude_functions,
        )



from __future__ import annotations

import threading
from collections import defaultdict
from dataclasses import dataclass
from enum import Enum
from typing import (
    Any,
    Dict,
    FrozenSet,
    List,
    NewType,
    Set,
    Type,
    TypeVar,
)

T = TypeVar("T")

ArgumentName = NewType("ArgumentName", str)


class ArgumentType(Enum):
    positional = "PositionalArgument"
    vararg = "VariableArgument"
    kwarg = "KeywordArgument"


Filename = NewType("Filename", str)
FunctionName = NewType("FunctionName", str)


@dataclass(eq=True, frozen=True)
class FuncInfo:
    file_name: Filename
    func_name: FunctionName


Typename = NewType("Typename", str)


@dataclass
class TypenameFrequency:
    typename: Typename
    counter: int

    def __hash__(self: TypenameFrequency) -> int:
        return hash(self.typename)

    def __eq__(self: TypenameFrequency, other: Any) -> bool:
        if isinstance(other, TypenameFrequency):
            return self.typename == other.typename
        return False


TypenameSet = NewType("TypenameSet", Set[TypenameFrequency])


@dataclass
class ArgInfo:
    arg_name: ArgumentName
    value_type: Type[Any]
    type_name_set: TypenameSet


@dataclass(eq=True, frozen=True)
class ImportDetails:
    object_name: str
    object_aliases: FrozenSet[str]
    import_module_name: str
    module_aliases: FrozenSet[str]


@dataclass(eq=True, frozen=True)
class ImportInfo:
    # 1. filename where the function lives
    function_fname: Filename
    # 2. filename where the class lives
    class_fname: Filename
    # 3. the name of the class
    class_name: str
    # 4. details for possible imports (see get_import_details).
    import_details: ImportDetails



import click
import functools
import inspect
import logging
import multiprocessing
import os
import random
import runpy
import signal
import sys
import time
import importlib.metadata

from collections import defaultdict
from types import (
    CodeType,
    FrameType,
    FunctionType,
)
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    ParamSpec,
    Set,
    TextIO,
    Tuple,
    TypeVar,
    get_type_hints,
)

from righttyper.get_import_details import (
    get_import_details,
)
from righttyper.righttyper_process import (
    collect_data,
    output_stub_files,
    process_file,
)
from righttyper.righttyper_runtime import (
    format_annotation,
    format_function_definition,
    get_adjusted_full_type,
    get_class_name_from_stack,
    get_class_source_file,
    requires_import,
    should_skip_function,
    update_argtypes,
)
from righttyper.righttyper_shapes import (
    print_annotation,
    update_arg_shapes,
    update_retval_shapes,
)
from righttyper.righttyper_tool import (
    register_monitoring_callbacks,
    reset_monitoring,
    setup_timer,
    setup_tool_id,
)
from righttyper.righttyper_types import (
    ArgInfo,
    ArgumentName,
    ArgumentType,
    Filename,
    FuncInfo,
    FunctionName,
    ImportInfo,
    Typename,
    TypenameFrequency,
    TypenameSet,
)
from righttyper.righttyper_utils import (
    TOOL_ID,
    TOOL_NAME,
    debug_print,
    debug_print_set_level,
    make_type_signature,
#    get_sampling_interval,
#    update_sampling_interval,
#    reset_sampling_interval,
    skip_this_file,
    unannotated,
    union_typeset_str,
)

# Below is to mollify mypy.
try:
    import sys.monitoring  # pyright: ignore
except Exception:
    pass

    
target_overhead : float = 5.0 # default
instrumentation_overhead = 0.0
alpha = 0.9
sample_count_instrumentation = 0.0
sample_count_total = 0.0

logger = logging.getLogger("righttyper")

# All visited functions (file name and function name)
visited_funcs: Set[FuncInfo] = set()

# Any function that yielded at least once (and so will be treated as a
# Generator)
yielded_funcs: Set[FuncInfo] = set()

# The existing spec for each function (that is, function prototype)
existing_spec: Dict[FuncInfo, str] = dict()

# For each visited function, all the info about its arguments
visited_funcs_arguments: Dict[FuncInfo, List[ArgInfo]] = dict()

# For each visited function, all the info about its return value
visited_funcs_retval: Dict[FuncInfo, TypenameSet] = defaultdict(
    lambda: TypenameSet(set())
)

# For each function and argument, what type the argument is (e.g.,
# kwarg, vararg)
arg_types: Dict[
    Tuple[FuncInfo, ArgumentName],
    ArgumentType,
] = dict()

# For each function, the variables (and, potentially, 'return') that
# have no type annotations
not_annotated: Dict[FuncInfo, Set[str]] = defaultdict(set)

# Existing annotations (variable to type annotations, optionally
# including 'return')
existing_annotations: Dict[FuncInfo, Dict[str, str]] = defaultdict(dict)

# Track the file names and class names for classes
# that will need import statements
imports: Set[ImportInfo] = set()


namespace: Dict[str, Any] = {}
script_dir = ""
include_files_regex = ""
include_all = False
infer_shapes = False # tensor shape inference

def enter_function(ignore_annotations: bool, code: CodeType) -> Any:
    """
    Process the function entry point, perform monitoring related operations,
    and manage the profiling of function execution.

    Args:
        code : CodeType object

    Returns:
        str: Status of monitoring
    """
    if should_skip_function(
        code,
        script_dir,
        include_all,
        include_files_regex,
    ):
        return sys.monitoring.DISABLE

    t = FuncInfo(
        Filename(code.co_filename),
        FunctionName(code.co_qualname),
    )
    
    visited_funcs.add(t)
    
    frame = inspect.currentframe()
    if frame and frame.f_back and frame.f_back.f_back:
        process_function_arguments(frame, t, ignore_annotations)

    return sys.monitoring.DISABLE


def call_handler(
    code: CodeType,
    instruction_offset: int,
    callable: object,
    arg0: object,
) -> Any:
    # If we are calling a function, activate its start, return, and yield handlers.
    if isinstance(callable, FunctionType):
        if should_skip_function(
            code,
            script_dir,
            include_all,
            include_files_regex,
        ):
            return sys.monitoring.DISABLE
        try:
            sys.monitoring.set_local_events(
                TOOL_ID,
                callable.__code__,
                sys.monitoring.events.PY_START
                | sys.monitoring.events.PY_RETURN
                | sys.monitoring.events.PY_YIELD,
            )
        except AttributeError:
            pass
    return sys.monitoring.DISABLE


def yield_function(
    code: CodeType,
    instruction_offset: int,
    return_value: Any,
) -> object:
    # We do the same thing for yields and exits.
    return exit_function_worker(
        code,
        instruction_offset,
        return_value,
        sys.monitoring.events.PY_YIELD,
    )


def exit_function(
    code: CodeType,
    instruction_offset: int,
    return_value: Any,
) -> object:
    return exit_function_worker(
        code,
        instruction_offset,
        return_value,
        sys.monitoring.events.PY_RETURN,
    )


def exit_function_worker(
    code: CodeType,
    instruction_offset: int,
    return_value: Any,
    event_type: int,
) -> object:
    """
    Function to gather statistics on a function call and determine
    whether it should be excluded from profiling, when the function exits.

    - If the function name is in the excluded list, it will disable the monitoring right away.
    - Otherwise, it calculates the execution time of the function, adds the type of the return value to a set for that function,
      and then disables the monitoring if appropriate.

    Args:
    code (CodeType): bytecode of the function.
    instruction_offset (int): position of the current instruction.
    return_value (Any): return value of the function.
    event_type (int): if this is a PY_RETURN (regular return) or a PY_YIELD (yield)

    Returns:
    int: indicator whether to continue the monitoring, always returns sys.monitoring.DISABLE in this function.
    """
    # Check if the function name is in the excluded list
    func_name = code.co_qualname
    filename = code.co_filename
    if should_skip_function(
        code,
        script_dir,
        include_all,
        include_files_regex,
    ):
        return sys.monitoring.DISABLE

    t = FuncInfo(
        Filename(filename),
        FunctionName(func_name),
    )

    # Special handling for functions that yielded.
    if t in yielded_funcs:
        # FIXME need to merge generator and return type for higher precision
        return sys.monitoring.DISABLE

    # Check if this is a method. If so, we need to replace anything using the class name with Self.
    class_name = get_class_name_from_stack()

    # Initialize if the function is first visited
    if t not in visited_funcs_retval:
        visited_funcs_retval[t] = TypenameSet(set())
    debug_print(f"exit processing, retval was {visited_funcs_retval[t]=}")

    if infer_shapes:
        update_retval_shapes(t, return_value)
    typename = get_adjusted_full_type(return_value, class_name)
    if event_type == sys.monitoring.events.PY_YIELD:
        # Yield: call it a generator
        # FIXME: We should be returning more precise Generators if we discover a return value.
        # See https://docs.python.org/3.10/library/typing.html#typing.Generator
        typename = f"Generator[{typename}, Any, Any]"
        yielded_funcs.add(t)

    # Check if the return value type is already in the set
    found = False
    for typename_frequency in visited_funcs_retval[t]:
        if typename_frequency.typename == typename:
            typename_frequency.counter += 1
            found = True
            break

    # If the return value type is not in the set, add it
    if not found:
        visited_funcs_retval[t].add(TypenameFrequency(Typename(typename), 1))

    return sys.monitoring.DISABLE


def process_function_arguments(
    frame: Any,
    t: FuncInfo,
    ignore_annotations: bool,
) -> None:
    # NOTE: this backtracking logic is brittle and must be
    # adjusted if the call chain increases in length.
    caller_frame = frame.f_back.f_back # .f_back
    code = caller_frame.f_code
    class_name = get_class_name_from_stack()
    args, varargs, varkw, the_values = inspect.getargvalues(
        caller_frame
    )
    if varargs:
        args.append(varargs)
    if varkw:
        args.append(varkw)

    type_hints = get_function_type_hints(
        caller_frame, code, ignore_annotations
    )
    if infer_shapes:
        update_arg_shapes(t, the_values)
    
    update_function_annotations(
        t,
        caller_frame,
        args,
        type_hints,
        ignore_annotations,
    )

    argtypes: List[ArgInfo] = []
    for arg in args:
        if arg:
            index = (
                FuncInfo(Filename(caller_frame.f_code.co_filename),
                         FunctionName(code.co_qualname)),
                ArgumentName(arg),
            )
            update_argtypes(
                argtypes,
                arg_types,
                index,
                the_values,
                class_name,
                arg,
                varargs,
                varkw,
            )

            if requires_import(the_values[arg]) and arg not in [
                "self",
                "cls",
            ]:
                # print(f"trying to add: {arg=} {the_values[arg]=}")
                try:
                    add_new_import(
                        caller_frame.f_code.co_filename,
                        the_values[arg],
                    )
                except TypeError:
                    # Unexpected type error
                    logger.exception(
                        "process_function_arguments - TypeError:"
                        f" {t=} {arg=} {the_values[arg]=}"
                    )

    debug_print(f"processing {t=} {argtypes=}")
    update_visited_funcs_arguments(t, argtypes)


@functools.cache
def get_function_type_hints(
    caller_frame: Any,
    code: CodeType,
    ignore_annotations: bool,
) -> Dict[str, str]:
    for (
        name,
        obj,
    ) in caller_frame.f_globals.items():
        if hasattr(obj, "__wrapped__"):
            obj = obj.__wrapped__
        if inspect.isfunction(obj) and obj.__code__ is code:
            try:
                return get_type_hints(obj)
            except Exception:
                return {}
    return {}


def update_function_annotations(
    t: FuncInfo,
    caller_frame: Any,
    args: List[str],
    type_hints: Dict[str, str],
    ignore_annotations: bool,
) -> None:
    for (
        name,
        obj,
    ) in caller_frame.f_globals.items():
        if hasattr(obj, "__wrapped__"):
            obj = obj.__wrapped__
        if inspect.isfunction(obj) and obj.__code__ is caller_frame.f_code:
            existing_spec[t] = format_function_definition(
                name,
                args,
                (type_hints if not ignore_annotations else {}),
            )
            not_annotated[t] = unannotated(obj, ignore_annotations)
            existing_annotations[t] = {
                name: format_annotation(type_hints[name])
                for name in type_hints
            }


def add_new_import(filename: str, value: Any) -> None:
    class_src_file = get_class_source_file(value.__class__)
    deets = get_import_details(value)
    new_import = ImportInfo(
        Filename(filename),
        Filename(class_src_file),
        value.__class__.__name__,
        deets,
    )
    imports.add(new_import)


def update_visited_funcs_arguments(
    t: FuncInfo, argtypes: List[ArgInfo]
) -> None:
    if t in visited_funcs_arguments:
        for i, arginfo in enumerate(argtypes):
            if i < len(visited_funcs_arguments[t]):
                update_argument_type(
                    visited_funcs_arguments[t][i],
                    arginfo.type_name_set,
                )
    else:
        visited_funcs_arguments[t] = argtypes


def update_argument_type(
    old_arginfo: ArgInfo,
    full_type_name_set: TypenameSet,
) -> None:
    for full_type_name in full_type_name_set:
        if any(
            full_type_name.typename == old_type_name.typename
            for old_type_name in old_arginfo.type_name_set
        ):
            for old_type_name in old_arginfo.type_name_set:
                if full_type_name.typename == old_type_name.typename:
                    old_type_name.counter += 1
                    break
        else:
            old_arginfo.type_name_set.add(next(iter(full_type_name_set)))
            # reset_sampling_interval()


instrumentation_functions = { 'enter_function', 'call_handler', 'exit_function_worker', 'restart_sampling' }

def restart_sampling(_signum: int, frame: Optional[FrameType]) -> None:
    """
    This function handles the task of clearing the seen functions.
    Called when a timer signal is received.

    Args:
        _signum: The signal number
        _frame: The current stack frame
    """
    # Walk the stack to see if righttyper instrumentation is running (and thus instrumentation).
    # We use this information to estimate instrumentation overhead, and put off restarting
    # instrumentation until overhead drops below the target threshold.
    global sample_count_instrumentation, sample_count_total
    global instrumentation_overhead
    sample_count_total += 1.0
    in_instrumentation = 0.0
    f = frame
    # We stop walking the stack after a given number of frames to
    # limit overhead. The instrumentation code should be fairly
    # shallow, so this heuristic should have no impact on accuracy
    # while improving performance.
    countdown = 10 
    while f and countdown > 0:
        func_name = f.f_code.co_qualname
        filename = f.f_code.co_filename
        if func_name in instrumentation_functions and 'righttyper.py' in filename:
            # In instrumentation code
            sample_count_instrumentation += 1.0
            break
        f = f.f_back
        countdown -= 1
    instrumentation_overhead = sample_count_instrumentation / sample_count_total
    if instrumentation_overhead <= target_overhead / 100.0:
        # Instrumentation overhead remains low enough; restart instrumentation.
        # Restart the system monitoring events
        sys.monitoring.restart_events()
    else:
        pass
    # Set a timer for the next round of sampling.
    signal.setitimer(
        signal.ITIMER_REAL,
        0.01,
    )


def output_type_signatures(
    file: TextIO = sys.stdout,
    namespace: Dict[str, Any] = globals(),
) -> None:
    # Print all type signatures
    fname_printed : Dict[Filename, bool] = defaultdict(bool)
    visited_funcs_by_fname = sorted(visited_funcs, key = lambda a: a.file_name + ":" + a.func_name)
    for t in visited_funcs_by_fname:
        if skip_this_file(
            t.file_name,
            script_dir,
            include_all,
            include_files_regex,
        ):
            continue
        try:
            s = make_type_signature(
                file_name=t.file_name,
                func_name=t.func_name,
                args=visited_funcs_arguments[t],
                retval=visited_funcs_retval[t],
                namespace=namespace,
                not_annotated=not_annotated,
                arg_types=arg_types,
                existing_annotations=existing_annotations,
            )
            if t in existing_spec and s == existing_spec[t]:
                continue
            if not fname_printed[t.file_name]:
                print(f"{t.file_name}:\n{'=' * (len(t.file_name) + 1)}\n", file=file)
                fname_printed[t.file_name] = True
            print(f"{s} ...\n", file=file)
            # Print diffs
            if t in existing_spec:
                import difflib
                diffs = difflib.ndiff((existing_spec[t] + "\n").splitlines(True),
                                      (s + "\n").splitlines(True))
                print(''.join(diffs), file=file)
            # First try at shapes
            annotations = print_annotation(t)
            try:
                ret_annotation = annotations.pop()
            except:
                ret_annotation = None
            if annotations and infer_shapes:
                # Process all annotations
                try:
                    annotations = [visited_funcs_arguments[t][index].arg_name + ": " + annotation.format(union_typeset_str(t.file_name, visited_funcs_arguments[t][index].type_name_set, {})) for index, annotation in enumerate(annotations)]
                    print("# Shape annotations", file=file)
                    print("@beartype", file=file)
                    if t in visited_funcs_retval:
                        assert ret_annotation
                        # Has a return value
                        retval_type = union_typeset_str(t.file_name, visited_funcs_retval[t], {})
                        print(f"def {t.func_name}({', '.join(annotations)}) -> {ret_annotation.format(retval_type)}: ...\n", file=file)
                    else:
                        print(f"def {t.func_name}({', '.join(annotations)}) -> None: ...\n", file=file)
                except IndexError:
                    # FIXME this should not happen, to track down later
                    logger.exception(f"IndexError in annotations")
                    
                    
        except KeyError:
            # Something weird happened
            logger.exception(f"KeyError: {t=}")


def initialize_globals(
    include_files: str,
    _include_all: bool,
    script: str,
    verbose: bool,
    _target_overhead: float,
    shapes_: bool
) -> None:
    debug_print_set_level(verbose)
    global include_files_regex, include_all, script_dir, target_overhead, infer_shapes
    include_files_regex = include_files
    include_all = _include_all
    script_dir = os.path.dirname(os.path.realpath(script))
    target_overhead = _target_overhead
    infer_shapes = shapes_

def execute_script_or_module(
    script: str,
    module: bool,
    tool_args: List[str],
    script_args: List[str],
) -> None:
    global namespace
    namespace = {}
    if module:
        sys.argv = [script] + tool_args
        try:
            namespace = runpy.run_module(
                script,
                run_name="__main__",
                alter_sys=True,
            )
        except SystemExit:
            pass
    else:
        sys.argv = [script] + script_args
        namespace = runpy.run_path(script, run_name="__main__")


def post_process(
    overwrite: bool = True,
    output_files: bool = True,
    ignore_annotations: bool = False,
    insert_imports: bool = False,
    generate_stubs: bool = False,
    srcdir: str = "",
) -> None:
    global namespace
    output_type_signatures_to_file(namespace)
    if generate_stubs:
        output_stub_files(
            namespace,
            srcdir,
            imports,
            visited_funcs,
            script_dir,
            include_all,
            include_files_regex,
            visited_funcs_arguments,
            visited_funcs_retval,
            not_annotated,
            arg_types,
            existing_annotations,
        )
    if output_files:
        process_all_files(
            ignore_annotations,
            overwrite,
            insert_imports,
            srcdir,
        )


def output_type_signatures_to_file(namespace: Dict[str, Any]) -> None:
    with open(f"{TOOL_NAME}.out", "w+") as f:
        output_type_signatures(f, namespace)


def process_all_files(
    ignore_annotations: bool,
    overwrite: bool,
    insert_imports: bool,
    srcdir: str,
) -> None:
    use_multiprocessing = True  # False

    processes: List[multiprocessing.Process] = []
    all_files = list(set(t.file_name for t in visited_funcs))
    prefix = os.path.commonprefix(list(all_files))

    # Collect the file names to process
    fnames_set = set()
    for t in visited_funcs:
        fname = t.file_name
        if fname and should_update_file(t, fname, prefix, namespace):
            assert not skip_this_file(
                fname,
                script_dir,
                include_all,
                include_files_regex,
            )
            fnames_set.add(fname)

    if len(fnames_set) == 0:
        return

    fnames = list(fnames_set)
    ### multiprocessing.set_start_method('fork')
    import rich.progress
    from rich.table import Column

    with rich.progress.Progress(  # rich.progress.TextColumn("[progress.description]{task.description}", table_column=Column(ratio=1)),
        rich.progress.BarColumn(table_column=Column(ratio=1)),
        rich.progress.MofNCompleteColumn(),
        rich.progress.TimeRemainingColumn(),
        transient=True,
        expand=True,
    ) as progress:
        task1 = progress.add_task(description="", total=len(fnames))
        for fname in fnames:
            type_annotations = collect_data(
                fname,
                visited_funcs,
                visited_funcs_arguments,
                visited_funcs_retval,
                namespace,
            )
            import_param = imports
            if not insert_imports:
                import_param = set()
            args = (
                fname,
                type_annotations,
                import_param,
                overwrite,
                not_annotated,
                ignore_annotations,
                srcdir,
            )
            if use_multiprocessing:
                process = multiprocessing.Process(
                    target=process_file, args=args
                )
                processes.append(process)
                process.start()
            else:
                process_file(*args)
                progress.update(task1, advance=1)

        if use_multiprocessing:
            while processes:
                for (
                    p
                ) in (
                    processes
                ):  # multiprocessing.connection.wait(processes, timeout=None):
                    p.join()
                    processes.remove(p)
                    progress.update(task1, advance=1)


def should_update_file(
    t: FuncInfo,
    fname: str,
    prefix: str,
    namespace: Dict[str, Any],
) -> bool:
    try:
        s = make_type_signature(
            file_name=t.file_name,
            func_name=t.func_name,
            args=visited_funcs_arguments[t],
            retval=visited_funcs_retval[t],
            namespace=namespace,
            not_annotated=not_annotated,
            arg_types=arg_types,
            existing_annotations=existing_annotations,
        )
    except KeyError:
        return False
    return t not in existing_spec or s != existing_spec[t]


FORMAT = "[%(filename)s:%(lineno)s] %(message)s"
logging.basicConfig(
    filename="righttyper.log",
    level=logging.INFO,
    format=FORMAT,
)
logger = logging.getLogger("righttyper")


class ScriptParamType(click.ParamType):
    def convert(self, value: str, param: Any, ctx: Any) -> str:
        # Check if it's a valid file path
        if os.path.isfile(value):
            return value
        # Check if it's an importable module
        try:
            importlib.import_module(value)
            return value
        except ImportError:
            self.fail(
                f"{value} is neither a file nor a module",
                param,
                ctx,
            )
            return ""


SCRIPT = ScriptParamType()


@click.command(
    context_settings=dict(
        ignore_unknown_options=True,
    )
)
@click.argument("script", type=SCRIPT, required=False)
@click.option(
    "--all-files",
    is_flag=True,
    help="Process any files encountered, including in libraries (except for those specified in --include-files)",
)
@click.option(
    "--include-files",
    type=str,
    help="Include only files matching the given regex pattern.",
)
@click.option(
    "--infer-shapes",
    is_flag=True,
    default=False,
    show_default=True,
    help="Produce tensor shape annotations (compatible with jaxtyping)."
)
@click.option(
    "--srcdir",
    type=click.Path(exists=True, file_okay=False),
    default=os.getcwd(),
    help="Use this directory as the base for imports.",
)
@click.option(
    "--overwrite/--no-overwrite",
    help="Overwrite files with type information.",
    default=False,
    show_default=True,
)
@click.option(
    "--output-files/--no-output-files",
    help="Output annotated files (possibly overwriting, if specified).",
    default=False,
    show_default=True,
)
@click.option(
    "--ignore-annotations",
    is_flag=True,
    help="Ignore existing annotations and overwrite with type information.",
    default=False,
)
@click.option(
    "-m",
    "--module",
    is_flag=True,
    help="Run the script as a module.",
)
@click.option(
    "--verbose",
    is_flag=True,
    help="Print diagnostic information.",
)
@click.option(
    "--insert-imports",
    is_flag=True,
    help="Insert import statements for missing classes (MAY LEAD TO CIRCULAR IMPORTS).",
    default=False,
)
@click.option(
    "--generate-stubs",
    is_flag=True,
    help="Generate stub files (.pyi).",
    default=False,
)
@click.option(
    "--type-coverage-by-directory",
    type=click.Path(exists=True, file_okay=False),
    help="Report per-directory type annotation coverage for all Python files in a directory and its children.",
)
@click.option(
    "--type-coverage-by-file",
    type=click.Path(exists=True, file_okay=False),
    help="Report per-file type annotation coverage for all Python files in a directory or its children.",
)
@click.option(
    "--type-coverage-summary",
    type=click.Path(exists=True, file_okay=False),
    help="Report uncovered and partially covered files and functions when performing type annotation coverage analysis.",
)  # Note: should only be available if coverage-by-directory or coverage-by-file are specified
@click.option(
    "---",
    "triple_dash",
    is_flag=True,
    help="Indicator to separate tool arguments from script or module arguments.",
    hidden=True,
)
@click.argument("args", nargs=-1, type=click.UNPROCESSED)
@click.version_option(
    version=importlib.metadata.version(TOOL_NAME),
    prog_name=TOOL_NAME,
)
@click.option(
    "--target-overhead",
    type=float,
    default=target_overhead,
    help="Target overhead, as a percentage (e.g., 5).",
)
def main(
    script: str,
    all_files: bool,
    include_files: str,
    module: bool,
    triple_dash: bool,
    args: List[str],
    type_coverage_by_directory: str,
    type_coverage_by_file: str,
    type_coverage_summary: str,
    verbose: bool,
    overwrite: bool,
    output_files: bool,
    ignore_annotations: bool,
    insert_imports: bool,
    generate_stubs: bool,
    infer_shapes: bool,
    srcdir: str,
    target_overhead: float,
) -> None:
    """
    RightTyper efficiently generates types for your function
    arguments and return values.
    """
    if infer_shapes:
        found = defaultdict(bool)
        try:
            import numpy as np
            found['numpy'] = True
            import pandas as pd
            found['pandas'] = True
            import torch
        except ModuleNotFoundError as e:
            print("At least one of the following packages need to be installed:")
            if not found['numpy']:
                print("  * numpy")
            if not found['pandas']:
                print("  * pandas")
            print("  * torch")
            sys.exit(1)
            
    if (
        type_coverage_by_directory or type_coverage_by_file
    ) and type_coverage_summary:
        raise click.UsageError(
            'The "--type-coverage-summary" option can only be specified when "--type-coverage-by-directory" or "--type-coverage-by-file" are NOT specified.'
        )

    from . import annotation_coverage

    if type_coverage_summary:
        directory_summary = annotation_coverage.analyze_all_directories(
            type_coverage_summary
        )
        annotation_coverage.print_annotation_summary()
        return

    if type_coverage_by_directory:
        directory_summary = annotation_coverage.analyze_all_directories(
            type_coverage_by_directory
        )
        annotation_coverage.print_directory_summary(directory_summary)
        return

    if type_coverage_by_file:
        file_summary = annotation_coverage.analyze_all_directories(
            type_coverage_by_file
        )
        annotation_coverage.print_file_summary(file_summary)
        return

    initialize_globals(include_files, all_files, script, verbose, target_overhead, infer_shapes)
    tool_args, script_args = split_args_at_triple_dash(args)
    setup_tool_id()
    register_monitoring_callbacks(
        enter_function,
        call_handler,
        exit_function,
        yield_function,
        ignore_annotations,
    )
    setup_timer(restart_sampling)
    execute_script_or_module(script, module, tool_args, script_args)
    reset_monitoring()
    post_process(
        overwrite=overwrite,
        output_files=output_files,
        ignore_annotations=ignore_annotations,
        insert_imports=insert_imports,
        generate_stubs=generate_stubs,
        srcdir=srcdir,
    )


def split_args_at_triple_dash(
    args: List[str],
) -> Tuple[List[str], List[str]]:
    tool_args = []
    script_args = []
    triple_dash_found = False
    for arg in args:
        if arg == "---":
            triple_dash_found = True
            continue
        if triple_dash_found:
            script_args.append(arg)
        else:
            tool_args.append(arg)
    return tool_args, script_args



from collections import defaultdict
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Set,
    Tuple,
)

from righttyper.righttyper_types import (
    FuncInfo
)

# FIXME needs to be thread local
current_shape : Dict[FuncInfo, List[Tuple[int, ...]]] = defaultdict(list)

captured_shapes : Dict[FuncInfo, Set[Tuple[int, ...]]] = defaultdict(set)

def transform_input(
    inp: List[Any],
) -> List[Tuple[Optional[str], ...]]:
    tup = inp[0]  # We assume all inputs have the same tuple lengths
    numargs = len(tup)
    all_vals: List[List[List[Optional[int]]]] = [
        [[] for j in range(len(tup[i]))] for i in range(numargs)
    ]
    output_vals: List[List[Optional[str]]] = [
        [None for j in range(len(tup[i]))] for i in range(numargs)
    ]
    hash_count: Dict[int, int] = defaultdict(int)
    hashes = [[0 for j in range(len(tup[i]))] for i in range(numargs)]
    # First, identify all the values that every shape dimension could hold;
    # then hash them all.
    for i in range(numargs):
        for j in range(len(tup[i])):
            v = []
            for input_entry in inp:
                try:
                    v.append(input_entry[i][j])
                except IndexError:
                    # FIXME - we should filter out the case where there are no shapes (all tuples are ()).
                    # print(f"{i=} {j=} {input_entry=}")
                    continue
            if len(set(v)) == 1:
                # All the same
                output_vals[i][j] = f"{v[0]}"
            all_vals[i][j] = v
            # Hash the entire column of values
            column_hash = hash(tuple(v))
            hash_count[column_hash] += 1
            hashes[i][j] = column_hash
    # Now post-process:
    # First, assign symbolic variables for each possible hash
    symbolic_variable = {}
    varindex = 0
    for k in hash_count:
        if hash_count[k] > 1:
            symbolic_variable[k] = f"dim{varindex}"
            varindex += 1
    # Now, replace every unassigned output value with its corresponding symbolic variable
    for i in range(numargs):
        for j in range(len(tup[i])):
            if not output_vals[i][j]:
                output_vals[i][j] = symbolic_variable[hashes[i][j]]
    output_tuples = [tuple(i) for i in output_vals]
    return output_tuples

def convert_to_jaxtyping(
    argument_datatypes: List[str],
    output_tuples: List[Tuple[Optional[str], ...]],
) -> List[str]:
    result = []
    for index, arg_tuple in enumerate(output_tuples):
        args = list(arg_tuple)
        s = []
        s_str = ""
        for arg in args:
            s.append(f"{arg}")
            s_str = " ".join(s)
        if s_str:
            declaration = f'{argument_datatypes[index]}[{{}}, "{s_str}"]'
        else:
            declaration = "{}"
        result.append(declaration)

    return result

def update_arg_shapes(func: FuncInfo, the_values: Dict[str, Any]) -> None:
    import numpy as np
    import pandas as pd
    import torch

    the_shapes : List[Any] = []
    for k in the_values:
        val = the_values[k]
        if isinstance(val, (pd.DataFrame, np.ndarray)):
            the_shapes.append(tuple(val.shape))
        elif isinstance(val, torch.Tensor):
            the_shapes.append(tuple(val.shape))
        else:
            the_shapes.append(tuple())
    current_shape[func].append(tuple(the_shapes))


def update_retval_shapes(func: FuncInfo, retval: Any) -> None:
    import numpy as np
    import pandas as pd
    import torch

    if len(current_shape[func]) == 0:
        return
    shapes = []
    if isinstance(retval, (pd.DataFrame, np.ndarray)):
        shapes.append(retval.shape)
    elif isinstance(retval, torch.Tensor):
        shapes.append(tuple(retval.shape))
    else:
        shapes.append(tuple())
    curr_shape = list(current_shape[func].pop())
    merged_shapes = tuple(curr_shape + shapes)
    captured_shapes[func].add(merged_shapes)
    # print(f"Shape for {func} = {merged_shapes}")
    
    
def print_annotation(func: FuncInfo) -> List[str]:
    # No annotations if the shape was never captured
    if func not in captured_shapes:
        return []
    # No annotations if all captured shapes are empty tuples
    if all(all(s == () for s in shape) for shape in list(captured_shapes[func])):
        return []
    tups = transform_input(list(captured_shapes[func]))
    n = len(tups)
    annotations = convert_to_jaxtyping(["Float" for i in range(n)], tups)
    return annotations


def print_annotations() -> None:
    for func in captured_shapes:
        print(f"{func}: {print_annotation(func)}")




import inspect
import random
import sys
from collections.abc import Generator
from functools import cache
from itertools import islice
from types import CodeType, ModuleType

from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
)

from righttyper.righttyper_types import (
    ArgInfo,
    ArgumentName,
    ArgumentType,
    FuncInfo,
    T,
    Typename,
    TypenameFrequency,
    TypenameSet,
)
from righttyper.righttyper_utils import (
    skip_this_file,
)


@cache
def should_skip_function(
    code: CodeType,
    script_dir: str,
    include_all: bool,
    include_files_regex: str,
) -> bool:
    if (
        code.co_name.startswith("<")
        or skip_this_file(
            code.co_filename,
            script_dir,
            include_all,
            include_files_regex,
        )
        or "righttyper" in code.co_filename
    ):
        return True
    if not (code.co_flags & 0x2):
        import dis

        assert dis.COMPILER_FLAG_NAMES[0x2] == "NEWLOCALS"
        return True
    return False


def get_class_name_from_stack(
    max_depth: int = 5,
) -> Optional[str]:
    # Initialize the current frame
    current_frame = inspect.currentframe()
    try:
        # Move up in the stack frame by frame
        depth = 0
        while current_frame and depth < max_depth:
            # Check if 'self' is in the local variables of the frame
            if "self" in current_frame.f_locals:
                instance = current_frame.f_locals["self"]
                return str(
                    instance.__class__.__name__
                )  # Return class name of the instance
            # Move one level up in the stack
            current_frame = current_frame.f_back
            depth += 1
    finally:
        # Delete reference to the current frame to avoid reference cycles
        del current_frame
    return None


def get_mypy_type_fn(func: Any) -> str:
    # Get the signature of the function
    signature = inspect.signature(func)

    # Extract argument types, default to Any if no annotation provided
    arg_types = [
        (param.annotation if param.annotation is not param.empty else Any)
        for name, param in signature.parameters.items()
    ]

    # Extract return type, default to Any if no annotation provided
    return_type = signature.return_annotation
    if return_type is inspect.Signature.empty:
        return_type = Any

    # Helper function to format the type names correctly
    def format_type(t: Any) -> Any:
        if hasattr(t, "__module__") and hasattr(t, "__qualname__"):
            return f"{t.__module__}.{t.__qualname__}"
        if hasattr(t, "__qualname__"):
            return t.__qualname__
        if hasattr(t, "__name__"):
            return t.__name__
        if hasattr(t, "_name"):  # Handling cases like Union
            return t._name
        if hasattr(t, "__origin__"):  # For generics
            return format_type(t.__origin__)
        return str(t)

    # Format the result
    arg_types_str = ", ".join([get_type_name(arg) for arg in arg_types])
    return_type_str = get_type_name(return_type)

    # Construct the Callable type string
    return f"Callable[[{arg_types_str}], {return_type_str}]"


def get_type_name(value: Any, depth: int = 0) -> str:
    # print(f"get_type_name {value}")
    orig_value = value
    if inspect.ismodule(value):
        # For modules, use their __name__ directly
        if getattr(value, "__name__", None):
            return value.__name__
        else:
            return str(type(value))
    elif not inspect.isclass(value):
        # If the value is an instance, get its class
        value = type(value)

    # print(f"HERE FOR {value}")
    # For types, use their __module__ and __name__
    module_name = value.__module__
    type_name = value.__name__

    if module_name == "builtins":
        if type_name == "frame":
            return "FrameType"
        if type_name == "function":
            return get_mypy_type_fn(orig_value)
        if type_name == "NoneType":
            return "None"
        if type_name == "list":
            return get_full_type(orig_value)
        if type_name == "tuple":
            return get_full_type(orig_value)
        if type_name == "Module":
            return get_full_type(orig_value)
        if type_name == "code":
            return "types.CodeType"
        return str(type_name)

    # Get the current global namespace
    current_namespace = sys._getframe(depth + 2).f_globals

    # Check if the type is accessible directly from the current namespace
    if (
        type_name in current_namespace
        and current_namespace[type_name] is value
    ):
        return str(type_name)

    # Check if the type is accessible from a module in the current namespace
    for name, obj in current_namespace.items():
        if (
            inspect.ismodule(obj)
            and hasattr(obj, type_name)
            and getattr(obj, type_name) is value
        ):
            return f"{name}.{type_name}"

    if module_name and module_name != "__main__":
        return f"{module_name}.{type_name}"
    else:
        return str(type_name)


def get_full_type(value: Any, depth: int = 0) -> str:
    """
    get_full_type function takes a value as input and returns a string representing the type of the value.

    If the value is of type dictionary, it randomly selects a pair of key and value from the dictionary
    and recursively determines their types.

    If the value is a list or a set, it randomly selects an item and determines its type recursively.

    If the value is a tuple, it determines the types of all elements in the tuple.

    For other types, it returns the name of the type.
    """
    if isinstance(value, dict):
        # Checking if the value is a dictionary
        if value:
            # If the dictionary is non-empty
            # we sample one of its items randomly.
            n = random.randint(0, len(value) - 1)
            # Here we are using islice with a starting position n and stopping at n + 1
            # to get a random key-value pair from the dictionary
            # FIXME: this is potentially costly and we should cap the range
            key, val = next(islice(value.items(), n, n + 1))
            # We return the type of the dictionary as 'dict[key_type: value_type]'
            return (
                f"Dict[{get_full_type(key, depth + 1)},"
                f" {get_full_type(val, depth + 1)}]"
            )
        else:
            # If the dictionary is empty, we just return 'dict' as the type
            return "Dict[Any]"
    elif isinstance(value, list):
        # Checking if the value is a list
        if value:
            # If the list is non-empty
            # we sample one of its elements randomly
            n = random.randint(0, len(value) - 1)
            elem = value[n]
            # We return the type of the list as 'list[element_type]'
            return f"List[{get_full_type(elem, depth + 1)}]"
        else:
            # If the list is empty, we return 'list'
            return "List[Any]"
    elif isinstance(value, tuple):
        if isinstance_namedtuple(value):
            return f"{value.__class__.__name__}"
        else:
            # Here we are returning the types of all elements in the tuple
            if len(value) == 0:
                tuple_str = "Tuple"
            else:
                tuple_str = (
                    f"Tuple[{', '.join(get_full_type(elem, depth + 1) for elem in value)}]"
                )
            return tuple_str
    elif inspect.ismethod(value):
        return get_mypy_type_fn(value)
    elif isinstance(value, set):
        if value:
            n = random.randint(0, len(value) - 1)
            elem = next(islice(value, n, n + 1))
            return f"Set[{get_full_type(elem, depth + 1)}]"
        else:
            return "Set[Any]"
    elif isinstance(value, Generator):
        # FIXME DISABLED FOR NOW
        # (q, g) = peek(value)
        # value = g
        # return f"Generator[{get_full_type(q)}, None, None]" # FIXME
        return "Generator[Any, None, None]"  # FIXME
    else:
        # If the value passed is not a dictionary, list, set, or tuple,
        # we return the type of the value as a string
        retval = get_type_name(value, depth)
        return retval


def get_adjusted_full_type(value: Any, class_name: Optional[str]) -> str:
    # Determine the type name of the return value
    if value is None:
        typename = "None"
    elif type(value) in (bool, float, int):
        typename = type(value).__name__
    else:
        typename = get_full_type(value)
        # print(f"typename = {typename}")
        # print(f"class name = {class_name}")
        if typename == class_name:
            typename = "Self"
        # print(f"typename now = {typename}")
    return typename


def isinstance_namedtuple(obj: object) -> bool:
    return (
        isinstance(obj, tuple)
        and hasattr(obj, "_asdict")
        and hasattr(obj, "_fields")
    )


def peek(
    generator: Generator[T, Any, Any],
) -> Tuple[T, Generator[T, Any, Any]]:
    def wrapped_generator() -> Generator[T, Any, Any]:
        # Yield the peeked value first, then yield from the original generator
        yield peeked_value
        yield from generator

    # Get the next value from the original generator
    peeked_value = next(generator)

    # Return the peeked value and the new generator
    return peeked_value, wrapped_generator()


def get_method_signature(method: Any) -> str:
    # try:
    #    type_hints = get_type_hints(method)
    # except Exception:
    #    type_hints = None
    callable_signature = get_mypy_type_fn(method)
    return callable_signature


def update_argtypes(
    argtypes: List[ArgInfo],
    arg_types: Dict[
        Tuple[FuncInfo, ArgumentName],
        ArgumentType,
    ],
    index: Tuple[FuncInfo, ArgumentName],
    the_values: Dict[str, Any],
    class_name: Optional[str],
    arg: str,
    varargs: Optional[str],
    varkw: Optional[str],
) -> None:

    def add_arg_info(
        argument_name: str,
        arg_type: type,
        values: Any,
        arg_type_enum: ArgumentType,
    ) -> None:
        types = TypenameSet({
            TypenameFrequency(
                Typename(get_adjusted_full_type(val, class_name)),
                1,
            )
            for val in values
        })
        argtypes.append(
            ArgInfo(
                ArgumentName(argument_name),
                arg_type,
                types,
            )
        )
        arg_types[index] = arg_type_enum

    if arg == varargs:
        assert varargs
        add_arg_info(
            varargs,
            tuple,
            the_values[arg],
            ArgumentType.vararg,
        )
    elif arg == varkw:
        assert varkw
        add_arg_info(
            varkw,
            dict,
            the_values[arg].values(),
            ArgumentType.kwarg,
        )
    else:
        add_arg_info(
            arg,
            type(the_values[arg]),
            [the_values[arg]],
            ArgumentType.positional,
        )


def format_annotation(annotation: Any) -> str:
    """Format an annotation (type hint) as a string."""
    if isinstance(annotation, type):
        return annotation.__name__
    elif hasattr(annotation, "_name") and annotation._name is not None:
        return str(annotation._name)
    elif (
        hasattr(annotation, "__origin__") and annotation.__origin__ is not None
    ):
        origin = format_annotation(annotation.__origin__)
        args = ", ".join(
            [format_annotation(arg) for arg in annotation.__args__]
        )
        return f"{origin}[{args}]"
    else:
        return str(annotation)


def format_function_definition(
    func_name: str,
    arg_names: List[str],
    type_hints: Dict[str, Any],
) -> str:
    """Format the function definition based on its name, argument names, and type hints."""
    params = []
    for arg in arg_names:
        type_hint = type_hints.get(arg)
        if type_hint:
            params.append(f"{arg}: {format_annotation(type_hint)}")
        else:
            params.append(arg)

    return_annotation = ""
    if "return" in type_hints:
        return_annotation = f" -> {format_annotation(type_hints['return'])}"

    params_str = ", ".join(params)
    function_definition = f"def {func_name}({params_str}){return_annotation}:"
    return function_definition


def get_class_source_file(cls: Type[Any]) -> str:
    module_name = cls.__module__

    # Check if the class is built-in
    if module_name == "builtins":
        return ""  # Built-in classes do not have source files

    try:
        # Try to get the module from sys.modules
        module = sys.modules[module_name]
        # Try to get the __file__ attribute
        file_path = getattr(module, "__file__", None)
        if file_path:
            return str(file_path)
        # If __file__ is not available, use inspect to get the source file
        import inspect

        file_path = inspect.getfile(cls)
        return file_path
    except (KeyError, TypeError, AttributeError):
        pass

    # Derive the file path from the module name
    try:
        # Assuming the module is part of the standard package structure
        import os

        module_parts = module_name.split(".")
        file_path = os.path.join(*module_parts)
        return file_path

    except Exception:
        pass

    return ""


def old_requires_import(obj: object) -> bool:
    # Get the class of the object
    cls = obj.__class__

    # Get the module where the class is defined
    cls_module = cls.__module__

    # Get the name of the current module
    current_module = sys._getframe(1).f_globals["__name__"]

    # If the class is defined in a different module, an import is needed
    if cls_module != current_module:
        return True
    return False


def requires_import(obj: object) -> bool:
    # Get the class of the object
    cls = type(obj)

    # Get the module where the class is defined
    cls_module = getattr(cls, "__module__", None)

    # If obj is a module or function from an imported module, use its module directly
    if isinstance(obj, ModuleType):
        cls_module = obj.__name__
    elif callable(obj):
        cls_module = getattr(obj, "__module__", None)

    # Handle cases where cls_module might be None
    if cls_module is None:
        return False

    # Get the name of the current module
    current_module = sys._getframe(4).f_globals.get("__name__", "")

    # Debug prints to see what's happening
    # print(f"Object: {obj}")
    # print(f"Class: {cls}")
    # print(f"Class module: {cls_module}")
    # print(f"Current module: {current_module}")

    # If the class is defined in a different module and not in builtins, an import is needed
    if cls_module != current_module and cls_module != "builtins":
        return True
    return False



import signal
import sys
from types import CodeType, FrameType
from typing import Any, Callable, Dict, Optional

from righttyper.righttyper_utils import (
    TOOL_ID,
    TOOL_NAME,
    get_sampling_interval,
)

_EVENTS = frozenset({
    sys.monitoring.events.PY_START,
    sys.monitoring.events.PY_RETURN,
    sys.monitoring.events.PY_YIELD,
    sys.monitoring.events.CALL,
})


def register_monitoring_callbacks(
    enter_function: Callable[[bool, CodeType], Any],
    call_handler: Callable[[CodeType, int, object, object], Any],
    exit_function: Callable[[CodeType, int, Any], object],
    yield_function: Callable[[CodeType, int, Any], object],
    ignore_annotations: bool,
) -> None:
    """Set up tracking for all enters, calls, exits, and yields."""
    event_set = 0
    for event in _EVENTS:
        event_set |= event

    sys.monitoring.set_events(TOOL_ID, event_set)

    fns : Dict[Any, Callable[..., Any]]  = {
        sys.monitoring.events.PY_START : (lambda x, y: enter_function(ignore_annotations, x)),
        sys.monitoring.events.CALL : call_handler,
        sys.monitoring.events.PY_RETURN : exit_function,
        sys.monitoring.events.PY_YIELD : yield_function,
    }

    for event in fns:
        sys.monitoring.register_callback(
            TOOL_ID,
            event,
            fns[event],
        )


def reset_monitoring() -> None:
    """Clear all monitoring of events."""
    for event in _EVENTS:
        sys.monitoring.register_callback(TOOL_ID, event, None)
    for id in range(3, 5):
        try:
            sys.monitoring.set_events(
                id,
                sys.monitoring.events.NO_EVENTS,
            )
        except ValueError:
            pass
    signal.signal(signal.SIGALRM, signal.SIG_IGN)
    signal.setitimer(signal.ITIMER_REAL, 0)


def setup_timer(
    func: Callable[[int, Optional[FrameType]], None],
) -> None:
    signal.signal(signal.SIGALRM, func)
    signal.setitimer(
        signal.ITIMER_REAL,
        0.01,
    )


def setup_tool_id() -> None:
    global TOOL_ID
    while True:
        try:
            sys.monitoring.use_tool_id(TOOL_ID, TOOL_NAME)
            break
        except Exception:
            TOOL_ID += 1



from typing import (
    Dict,
    List,
    Optional,
    Set,
    Tuple,
    Union,
)

import libcst as cst

from righttyper.righttyper_types import (
    ArgumentName,
    Filename,
    FuncInfo,
    FunctionName,
    Typename,
)


class AnnotateFunctionTransformer(cst.CSTTransformer):
    def __init__(
        self,
        filename: str,
        type_annotations: Dict[
            FuncInfo,
            Tuple[
                List[Tuple[ArgumentName, Typename]],
                Typename,
            ],
        ],
        not_annotated: Dict[FuncInfo, Set[str]],
    ) -> None:
        self.filename = filename
        self.type_annotations = type_annotations
        self.not_annotated = not_annotated
        self.class_name: List[str] = []

    def visit_ClassDef(self, node: cst.ClassDef) -> Optional[bool]:
        self.class_name.append(node.name.value)
        return None

    def leave_ClassDef(
        self,
        original_node: cst.ClassDef,
        updated_node: cst.ClassDef,
    ) -> Union[cst.BaseStatement, cst.RemovalSentinel]:
        self.class_name.pop()
        return original_node

    def leave_FunctionDef(
        self,
        original_node: cst.FunctionDef,
        updated_node: cst.FunctionDef,
    ) -> cst.FunctionDef:
        # real_name = ".".join(self.class_name + [original_node.name.value])
        name = original_node.name.value

        # TODO: the current handling will lead to a bug if a function
        # appears inside and outside a class definition in the same
        # file. We need to handle this case.

        key = FuncInfo(
            Filename(self.filename),
            FunctionName(name),
        )  # original_node.name.value))
        if key in self.type_annotations:
            args, return_type = self.type_annotations[key]
            # print(f"{args=}, {return_type=}")
            # Update arguments with type annotations
            new_parameters = []
            for parameter in updated_node.params.params:
                for arg, annotation_ in args:
                    if parameter.name.value == arg:
                        if arg not in self.not_annotated[key]:
                            continue
                        parsed_annotation = cst.parse_expression(annotation_)
                        new_parameters.append(
                            parameter.with_changes(
                                annotation=cst.Annotation(
                                    annotation=parsed_annotation
                                )
                            )
                        )
                        break
                else:
                    new_parameters.append(parameter)
            parsed_return_type = cst.parse_expression(return_type)
            if "return" in self.not_annotated[key]:
                updated_node = updated_node.with_changes(
                    params=updated_node.params.with_changes(
                        params=new_parameters
                    ),
                    returns=cst.Annotation(annotation=parsed_return_type),
                )
            else:
                updated_node = updated_node.with_changes(
                    params=updated_node.params.with_changes(
                        params=new_parameters
                    ),
                )
        return updated_node
